#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=
'''
Handling and creation of ICC profiles in an OO front end to LCMS bindings.
'''

## Created: 2009-01-17 Guy K. Kloss <G.Kloss@massey.ac.nz>
## Changed: $Id:$
##
## Copyright 2009 Guy K. Kloss
## All rights reserved.
##
## This software is distributed under the terms of the
## GNU Lesser General Public License
##
## http://www.massey.ac.nz/~gkloss/

__author__ = 'Guy K. Kloss <G.Kloss@massey.ac.nz>'
__date__ = '$Date: $'
__version__ = '$Revision: $'[11:-2]

import os
from littlecms import c_lcms, tag
from littlecms.general import LcmsException
from littlecms.support_structures import CieColour
from littlecms.tag import Tag


def _tagFactory(tagType, payload=None, profile=None):
    """
    Produces new tag objects depending on the tag type. If a payload or a
    profile is given it tries to initialise its payload from these.
    
    @param tagType: ICC tagType of tag to create.
    @tagType tagType: C{int}
    @param payload: "Simple" payload of the tag (default: None).
        Usually a string or number.
    @type payload: C{object}
    @param profile: A profile to extract the tag information from.
    @type profile: L{Profile}
    
    @return: A tag object.
    @rtype: A object of the interface type L{Tag}.
    """
    tag_type = tag.TAG_CLASS_MAP[tagType]
    if tagType not in tag_type.SUPPORTED_TAGS:
        tag_type = Tag
    if payload is not None:
        return tag_type(tagType, payload=payload)
    elif profile is not None:
        if tagType in tag_type.SUPPORTED_TAGS:
            return tag_type(tagType, profile=profile)
        return Tag(tagType)
    else:
        return tag_type(tagType)
    


class Profile(object):
    """
    Profile handling class.
    """

    def __init__(self,
                 fileName=None,
                 colourSpace=None,
                 lcmsProfile=None,
                 rawProfile=None,
                 whitePoint=None):
        """
        The constructor, opening a profile for reading.

        @param fileName: File name of an ICC profile to load (default: None).
        @type fileName: C{string}
        @param colourSpace: Type of colour space used for creating profiles from
            LCMS built in profiles (default: None). Refer to the LCMS API 
            documentation at U{http://littlecms.com/LCMSAPI.TXT} for valid 
            colour spaces. Supported colour spaces are L{PT_RGB} (for sRGB),
            L{PT_Lab} and L{PT_XYZ).
        @type colourSpace: C{int}
        @param lcmsProfile: A "raw" L{c_lcms} profile (default: None). It will
            be disposed upon deletion of this object.
        @type lcmsProfile: L{c_lcms.cmsHPROFILE}
        @param rawProfile: A "raw" L{c_lcms} profile (default: None). The same
            as lcmsProfile, but it will not be disposed upon deletion of this
            object.
        @type rawProfile: L{c_lcms.cmsHPROFILE}
        @param whitePoint: Only together with creating a built in Lab profile
            with the colourSpace=L{PT_Lab} option (default: None).
        @type whitePoint: L{CieColour}
        """
        self._profile = None
        self._rawProfile = False
        self.whitePoint = CieColour([c_lcms.D50X, c_lcms.D50Y, c_lcms.D50Z],
                                    'XYZ')
        if whitePoint != None:
            if isinstance(whitePoint, CieColour):
                self.whitePoint = whitePoint
            elif isinstance(whitePoint, c_lcms.cmsCIExyY):
                self.whitePoint = CieColour(whitePoint)
            else:
                raise LcmsException('Unsupported colour description for white point.')
                    
        if rawProfile:
            self._rawProfile = True
      #      self._profile = rawProfile
            self._profile = c_lcms.cmsOpenProfileFromMem(rawProfile[0], rawProfile[1])
        elif lcmsProfile:
            self._profile = lcmsProfile
        elif fileName:
            if os.path.isfile(fileName):
                self._profile = c_lcms.cmsOpenProfileFromFile(fileName, 'r')
            else:
                raise IOError("No such file or directory: '%s'" % fileName)
        elif colourSpace == c_lcms.PT_Lab:
            if whitePoint != None:
                self._profile = c_lcms.cmsCreateLab4Profile(self.whitePoint.lcms_xyY)
            else:
                self._profile = c_lcms.cmsCreateLab4Profile(None)
        elif colourSpace == c_lcms.PT_XYZ:
            self._profile = c_lcms.cmsCreateXYZProfile()
        elif colourSpace == c_lcms.PT_RGB:
            self._profile = c_lcms.cmsCreate_sRGBProfile()
        else:
            raise LcmsException('Unknown profile type to create.')
        
        self._tags = None


    def __del__(self):
        """
        Auto-closing profile.
        """
        if self._profile and not self._rawProfile:
            c_lcms.cmsCloseProfile(self._profile)


    @property
    def name(self):
        """
        Get the ICC profile "name" attribute.
        """
        return c_lcms.cmsTakeProductName(self._profile)


    @property
    def colourSpace(self):
        """
        Get the ICC profile "name" attribute.
        """
        return c_lcms.COLOUR_TYPE_FROM_ICC[c_lcms.cmsGetColorSpace(self._profile)]


    @property
    def lcmsProfile(self):
        """
        Get the LCMS profile object.
        """
        return self._profile


    def setTag(self, a_tag):
        """
        Set a tag in current tag list.
        
        @param a_tag: Tag to set.
        @type a_tag: L{littlecms.Tag}
        """
        _ = self.tags
        self._tags[a_tag.iccTagType] = a_tag
        
    
    @property
    def tags(self):
        """
        Accesses all tags of a profile, and adds them to a dictionary.
        
        @return: All tags, indexed by the tag signature, using L{littlecms.Tag}
            objects as values.
        @rtype: C{dict} 
        """
        if self._tags is None:
            self._tags = {}
            number_of_tags = c_lcms.cmsGetTagCount(self._profile)
            print number_of_tags
            for i in range(number_of_tags):
                a_tag = c_lcms.cmsGetTagSignature(self._profile, i)
                try:
                    self._tags[a_tag] = _tagFactory(a_tag, profile=self)
                except:pass
        return self._tags
    
    
    def save(self, filename):
        """
        Save the profile to a file.
        
        @param filename: Name of the file to write the profile to.
        @type filename: C{string}
        """
        for tag_sig, a_tag in self._tags.items():
            c_lcms.cmsAddTag(self._profile, tag_sig, a_tag.lcms_payload)
        c_lcms.cmsSaveProfile(self._profile, filename)
        
    
    @property
    def mediaWhitePoint(self):
        """
        Get the media white point.
        """
        media_white = c_lcms.cmsCIEXYZ(0, 0, 0)
        c_lcms.cmsTakeMediaWhitePoint(media_white, self._profile)
        return CieColour(media_white)

    @property
    def illuminant(self):
        illuminant = c_lcms.cmsCIEXYZ(0, 0, 0)
        c_lcms.cmsTakeIluminant(illuminant, self._profile)
        return CieColour(illuminant)
        
    @property
    def productName(self):
        """
        Get the stored product name.
        """
        return c_lcms.cmsTakeProductName(self._profile)

    @property
    def productInfo(self):
        """
        Get the stored product name.
        """
        return c_lcms.cmsTakeProductInfo(self._profile)

    @property
    def productDesc(self):
        """
        Get the stored product name.
        """
        return c_lcms.cmsTakeProductDesc(self._profile)
    
