# -*- coding: utf-8 -*-
'''Supporting structures that shadow one or more low level LCMS structures.'''

## Created: 2009-11-21 Guy K. Kloss <G.Kloss@massey.ac.nz>
## Changed: $Id:$
##
## Copyright 2009 Guy K. Kloss
## All rights reserved.
##
## http://www.massey.ac.nz/~gkloss/

__author__ = 'Guy K. Kloss <G.Kloss@massey.ac.nz>'
__date__ = '$Date: $'
__version__ = '$Revision: $'[11:-2]

import numpy
import ctypes
from operator import attrgetter
from littlecms import c_lcms
from littlecms.general import LcmsException


class CieColour(object):
    """
    Handles CIE colour descriptors in the different CIE formats. All colour
    space tuples are assumed to be interpreted at D50.
    """
    def __init__(self, colour=None, colourSpace=''):
        # The ground truth colour storage is as a floating point colour tuple
        # in CIE XYZ colour space.
        self._array_XYZ = numpy.zeros(3, dtype=float)
        if colour is not None:
            if isinstance(colour, numpy.ndarray) or \
                isinstance(colour, list) or \
                isinstance(colour, tuple):
                self._arraySet(colour, colourSpace)
            elif isinstance(colour, CieColour):
                self._array_XYZ[:] = colour.XYZ
            else:
                self._lcmsSet(colour)
        
        
    def __repr__(self):
        return "%s(%s, 'XYZ')" % (self.__class__.__name__, self._array_XYZ.tolist())
        

    def get(self, colourSpace):
        """
        Returns the colour in the requested colour space as an array (3 tuple).
        
        @param colourSpace: The requested destination colour space (one of 'XYZ',
            'Lab', 'LCh' or 'xyY').
        @type colourSpace: C{string}
        
        @return: Colour tuple.
        @rtype: L{numpy.ndarray} with 3 elements
        """
        if colourSpace == 'XYZ':
            return self._array_XYZ
        elif colourSpace == 'Lab':
            lab = self.lcmsGet(colourSpace)
            return numpy.array([lab.L, lab.a, lab.b], dtype=float)
        elif colourSpace == 'xyY':
            xyY = self.lcmsGet(colourSpace)
            return numpy.array([xyY.x, xyY.y, xyY.Y], dtype=float)
        elif colourSpace == 'LCh':
            lch = self.lcmsGet(colourSpace)
            return numpy.array([lch.L, lch.C, lch.h], dtype=float)
        else:
            raise LcmsException('Unsupported colour space: "%s".' % colourSpace)


    def lcmsGet(self, colourSpace):
        """
        Returns the colour in the requested colour space as a native LCMS
        colour type.
        
        @param colourSpace: The requested destination colour space (one of 'XYZ',
            'Lab', 'LCh' or 'xyY').
        @type colourSpace: C{string}
        
        @return: LCMS colour structure.
        @rtype: L{cmsCIEXYZ}, L{cmsCIELab} or L{cmsCIExyY}
        """
        if colourSpace == 'XYZ':
            return c_lcms.cmsCIEXYZ(*self._array_XYZ) # pylint: disable-msg=W0142
        elif colourSpace == 'Lab':
            xyz = self.lcmsGet('XYZ')
            lab = c_lcms.cmsCIELab()
            c_lcms.cmsXYZ2Lab(None, lab, xyz)
            return lab
        elif colourSpace == 'xyY':
            xyz = self.lcmsGet('XYZ')
            xyY = c_lcms.cmsCIExyY()
            c_lcms.cmsXYZ2xyY(xyY, xyz)
            return xyY
        elif colourSpace == 'LCh':
            lab = self.lcmsGet('Lab')
            lch = c_lcms.cmsCIELCh()
            c_lcms.cmsLab2LCh(lch, lab)
            return lch
        else:
            raise LcmsException('Unsupported colour space: "%s".' % colourSpace)


    def set(self, colour, colourSpace=None):
        """
        Sets the colour from the requested colour space.
        
        @param colour: LCMS colour structure to set.
        @type colour: L{cmsCIEXYZ}, L{cmsCIELab}, L{cmsCIExyY} or Python sequence
            with 3 elements (including L{numpy.ndarray}).
        @param colourSpace: The requested destination colour space (one of 'XYZ',
            'Lab', 'LCh' or 'xyY'). For LCMS colour types this parameter can be
            omitted.
        @type colourSpace: C{string}
        """
        if isinstance(colour, numpy.ndarray) or \
            isinstance(colour, list) or \
            isinstance(colour, tuple):
            self._arraySet(colour, colourSpace)
        else:
            self._lcmsSet(colour)

        
    def _lcmsSet(self, colour):
        """
        Sets the colour from the requested colour space in a native LCMS
        colour type.
        
        @param colour: LCMS colour structure to set.
        @type colour: L{cmsCIEXYZ}, L{cmsCIELab}, L{cmsCIELCh} or L{cmsCIExyY}
        """
        xyz = None
        if isinstance(colour, c_lcms.cmsCIEXYZ):
            xyz = colour
        elif isinstance(colour, c_lcms.cmsCIELab):
            xyz = c_lcms.cmsCIEXYZ()
            c_lcms.cmsLab2XYZ(None, xyz, colour)
        elif isinstance(colour, c_lcms.cmsCIExyY):
            xyz = c_lcms.cmsCIEXYZ()
            c_lcms.cmsxyY2XYZ(xyz, colour)
        elif isinstance(colour, c_lcms.cmsCIELCh):
            lab = c_lcms.cmsCIELab()
            c_lcms.cmsLCh2Lab(lab, colour)
            xyz = c_lcms.cmsCIEXYZ()
            c_lcms.cmsLab2XYZ(None, xyz, lab)
        else:
            raise LcmsException('Unsupported colour space descriptor: "%s".' % colour)

        self._array_XYZ[0] = xyz.X
        self._array_XYZ[1] = xyz.Y
        self._array_XYZ[2] = xyz.Z


    def _arraySet(self, colour, colourSpace):
        """
        Sets the colour from the requested colour space in an array (3 tuple).
        
        @param colour: Colour tuple.
        @type colour: L{numpy.ndarray} with 3 elements
        @param colourSpace: The requested destination colour space (one of 'XYZ',
            'Lab' or 'xyY').
        @type colourSpace: C{string}
        """
        if colourSpace == 'XYZ':
            self._array_XYZ[:] = colour
        elif colourSpace == 'Lab':
            lab = c_lcms.cmsCIELab(*colour) # pylint: disable-msg=W0142
            self._lcmsSet(lab)
        elif colourSpace == 'xyY':
            xyY = c_lcms.cmsCIExyY(*colour) # pylint: disable-msg=W0142
            self._lcmsSet(xyY)
        elif colourSpace == 'LCh':
            lab = c_lcms.cmsCIELCh(*colour) # pylint: disable-msg=W0142
            self._lcmsSet(lab)
        else:
            raise LcmsException('Unsupported colour space: "%s".' % colourSpace)


    def _set_XYZ(self, colour):
        self.set(colour, 'XYZ')

    XYZ = property(attrgetter('_array_XYZ'), _set_XYZ)
    

    def _get_Lab_asArray(self):
        return self.get('Lab')

    def _set_Lab(self, colour):
        self._arraySet(colour, 'Lab')
        
    Lab = property(_get_Lab_asArray, _set_Lab)


    def _get_LCh_asArray(self):
        return self.get('LCh')

    def _set_LCh(self, colour):
        self._arraySet(colour, 'LCh')
        
    LCh = property(_get_LCh_asArray, _set_LCh)


    def _get_xyY_asArray(self):
        return self.get('xyY')

    def _set_xyY(self, colour):
        self._arraySet(colour, 'xyY')
        
    xyY = property(_get_xyY_asArray, _set_xyY)


    def _get_XYZ_asLcmsType(self):
        return self.lcmsGet('XYZ')
    
    lcms_XYZ = property(_get_XYZ_asLcmsType, _set_XYZ)
    
    
    def _get_Lab_asLcmsType(self):
        return self.lcmsGet('Lab')
    
    lcms_Lab = property(_get_Lab_asLcmsType, _set_Lab)
    
    
    def _get_LCh_asLcmsType(self):
        return self.lcmsGet('LCh')
    
    lcms_LCh = property(_get_LCh_asLcmsType, _set_LCh)
    
    
    def _get_xyY_asLcmsType(self):
        return self.lcmsGet('xyY')
    
    lcms_xyY = property(_get_xyY_asLcmsType, _set_xyY)
    


class LUT(object):
    """
    Colour Look Up Table (CLUT) handling class.
    """

    def __init__(self, resolution=17, inputChannels=3, outputChannels=3,
                 lcms_lut=None):
        """
        Constructor.
        
        @param resolution: Number of nodes per side of the LUT's hyper cube
            (default: 17).
        @tagType resolution: C{int}
        @param inputChannels: Number of colour channels on source side of the
            LUT (default: 3).
        @type inputChannels: C{int}
        @param outputChannels: Number of colour channels on destination side of
            the LUT (default: 3.
        @type outputChannels: C{int}
        @param lcms_lut: LCMS LUT structure to initialise this LUT object from.
        @type lcms_lut: L{littlecms.c_lcms.LPLUT}
        """
        self.resolution = 0
        self.inputChannels = 0
        self.outputChannels = 0
        self._lcms_LUT = None

        if lcms_lut is not None:
            self._lcms_LUT = lcms_lut
            self.resolution = lcms_lut[0].cLutPoints
            self.inputChannels = lcms_lut[0].InputChan
            self.outputChannels = lcms_lut[0].OutputChan
        else:
            self._lcms_LUT = c_lcms.cmsAllocLUT()
            self.resolution = resolution
            self.inputChannels = inputChannels
            self.outputChannels = outputChannels
            c_lcms.cmsAlloc3DGrid(self._lcms_LUT, self.resolution,
                                  self.inputChannels, self.outputChannels)
    
        lut_shape = [self.resolution] * self.inputChannels + [self.outputChannels]
        buffer_elements = self.resolution ** self.inputChannels * self.outputChannels
        LutMatrixBufferType = c_lcms.WORD * buffer_elements # pylint: disable-msg=C0103
        ctypes_array = LutMatrixBufferType.from_address(ctypes.addressof(self._lcms_LUT[0].T.contents))
        self._array = numpy.ctypeslib.as_array(ctypes_array)
        self._array.shape = lut_shape


    def __del__(self):
        c_lcms.cmsFreeLUT(self._lcms_LUT)
        

    array = property(attrgetter('_array'))
    content = property(attrgetter('_lcms_LUT'))

    
    def callbackFunction(self, inVector, outVector): # pylint: disable-msg=R0201
        """
        Callback function for grid sampling of the LUT hyper cube. This
        function is called for every node in the LUT. This implementation is a
        dummy as it will copy the C{inVector} to the C{outVector} and therefore
        resembles an identity transformation.
        
        @param inVector: Coordinates of LUT node in 16 bit unsigned coordinates.
            A node index 0 is therefore 0x0000, and the last index of each
            dimension is 0xffff.
        @type inVector: L{numpy.ndarray} with C{ctype} L{numpy.uint16}
        @param outVector: Look up value of LUT node as an unsigned 16 bit
            integer.
        @type outVector: L{numpy.ndarray} with C{ctype} L{numpy.uint16}
        
        @return: The out vector.
        @rtype: L{numpy.ndarray} with C{ctype} L{numpy.uint16}
        """
        outVector[:] = inVector
        return outVector


    def _increment_index(self, index):
        """
        Internal helper to determine the next higher hyper cube index of the
        flattened array structure.
        
        @param index: Coordinate index of the hyper cube with C{inputChannels}
            elements in the index vector.
        @type index: L{numpy.ndarray} of type C{int}
        """
        position = len(index)
        while position >= 0:
            position -= 1
            if index[position] < self.resolution - 1:
                index[position] += 1
                break
            else:
                index[position] = 0

    
    def sampleGrid(self, callbackFunction=None):
        """
        Sampling of the grid nodes in the LUT hyper cube. The callback function
        takes two arguments, an C{inVector} and an C{outVector}, and it returns
        also the C{outVector}. The implementation has to match formally the
        description of L{callbackFunction}.
        
        @param callbackFunction: Callback function (default: None). If not
            given, the object's L{callbackFunction} method will be used.
        @type callbackFunction: Callable of type L{callbackFunction}
        """
        if not callbackFunction:
            callbackFunction = self.callbackFunction
        
        # Use a flattened hyper cube to iterate over all grid nodes.
        matrix_shape = self._array.shape
        grid_nodes = self.resolution ** self.inputChannels
        self._array.shape = (grid_nodes, self.outputChannels)
        
        # Populate the LUT array.
        scale = float(0xffff) / (self.resolution - 1)
        index = numpy.zeros(self.inputChannels)
        in_vector = numpy.zeros(self.inputChannels, dtype=numpy.uint16)
        for i in xrange(grid_nodes):
            in_vector[:] = numpy.round(scale * index)
            callbackFunction(in_vector, self._array[i]) 
            self._increment_index(index)
        
        # Return to our old shape.
        self._array.shape = matrix_shape
