#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=
'''
Handling and creation of ICC profile tags in an OO front end to LCMS bindings.
'''

## Created: 2009-01-19 Guy K. Kloss <G.Kloss@massey.ac.nz>
## Changed: $Id:$
##
## Copyright 2007 Guy K. Kloss
## All rights reserved.
##
## This software is distributed under the terms of the
## GNU Lesser General Public License
##
## http://www.massey.ac.nz/~gkloss/

__author__ = 'Guy K. Kloss <G.Kloss@massey.ac.nz>'
__date__ = '$Date: $'
__version__ = '$Revision: $'[11:-2]

import ctypes
from operator import attrgetter
from littlecms import c_lcms, tagSignature2text
from littlecms.general import LcmsException
from littlecms.support_structures import CieColour, LUT

_ALL_ICC_TAG_NAMES = [x for x in dir(c_lcms)
                      if x.startswith('icSig') and x.endswith('Tag')]
_ALL_ICC_TAGS = [getattr(c_lcms, x) for x in _ALL_ICC_TAG_NAMES]
ICC_TAG_MAP = dict(zip(_ALL_ICC_TAGS, _ALL_ICC_TAG_NAMES))


class Tag(object):
    """
    Class for (generic) profile tags. 
    """
    ALLOWED_TAGS = ()
    SUPPORTED_TAGS = ()

    
    def __init__(self, tagType, payload=None, profile=None):
        """
        Constructor.
        
        @param tagType: ICC tagType of tag to create.
        @tagType tagType: C{int}
        @param payload: Payload of the tag (default: None).
            This is in terms that the low level L{c_lcms} bindings can
            understand directly.
        @type payload: C{object}
        @param profile: A profile to extract the tag information from.
        @type profile: L{Profile}
        """
        self.iccTagType = tagType
        if tagType not in ICC_TAG_MAP:
            raise LcmsException('Unknown ICC tag type: %d' % tagType)
        self.payload = payload
        
        # We want to be able to instantiate this class for all tag types,
        # but if we're inherited, enforce the ALLOWED and SUPPORTED tags. 
        if type(self) is not Tag:
            # Refuse anything but allowed tags.
            if tagType not in self.ALLOWED_TAGS:
                raise LcmsException('Not allowed ICC tag type for %s: %s (%d)'
                                    % (self.__class__.__name__,
                                       ICC_TAG_MAP[tagType], tagType))
            # We can only do these from known and good payloads.
            elif tagType not in self.SUPPORTED_TAGS and payload is None:
                raise LcmsException('Unsupported ICC tag type for %s: %s (%d)'
                                    % (self.__class__.__name__,
                                       ICC_TAG_MAP[tagType], tagType))
        
        if profile is not None:
            self._content_from_profile(tagType, profile)
            
        
    def __str__(self):
        return "%s(tagType='%s')" % (self.__class__.__name__,
                                     ICC_TAG_MAP[self.iccTagType])


    lcms_payload = property(attrgetter('payload'))


    @property
    def iccTagTypeString(self):
        """
        Returns the tag type as a four character string as used in the
        ICC profile itself.
        
        @return: Four character ICC tag string.
        @rtype: C{string}
        """
        return tagSignature2text(self.iccTagType)
    
    
    @property
    def iccTagTypeDescription(self):
        """
        Returns the tag type as a descriptive string as used in the
        ICC header files.
        
        @return: Tag description as used in the C header files.
        @rtype: C{string}
        """
        return ICC_TAG_MAP[self.iccTagType]
    
    
    def _content_from_profile(self, tagType, profile):
        """Extract tag information from profile."""
        if tagType not in self.SUPPORTED_TAGS:
            raise LcmsException('Unsupported ICC tag type for %s: %s (%d)'
                                % (self.__class__.__name__,
                                   ICC_TAG_MAP[tagType], tagType))



class TextTag(Tag):
    """
    Text type tags.
    """
    ALLOWED_TAGS = (c_lcms.icSigCharTargetTag,
                    c_lcms.icSigCopyrightTag,
                    c_lcms.icSigProfileDescriptionTag,
                    c_lcms.icSigDeviceMfgDescTag,
                    c_lcms.icSigDeviceModelDescTag)
    SUPPORTED_TAGS = (c_lcms.icSigCopyrightTag,
                      c_lcms.icSigProfileDescriptionTag,
                      c_lcms.icSigDeviceMfgDescTag,
                      c_lcms.icSigDeviceModelDescTag)
    _ACCESSOR_FUNCTIONS = {c_lcms.icSigCopyrightTag:          c_lcms.cmsTakeCopyright,
                           c_lcms.icSigProfileDescriptionTag: c_lcms.cmsTakeProductDesc,
                           c_lcms.icSigDeviceMfgDescTag:      c_lcms.cmsTakeManufacturer,
                           c_lcms.icSigDeviceModelDescTag:    c_lcms.cmsTakeModel}

    
    def _content_from_profile(self, tagType, profile):
        """Extract tag information from profile."""
        Tag._content_from_profile(self, tagType, profile)
    
        accessor_function = self._ACCESSOR_FUNCTIONS[tagType]
        self.payload = accessor_function(profile.lcmsProfile)
        
       

class LutTag(Tag):
    """
    Colour Look Up Table (CLUT) type tags.
    """
    # CLUT content is in c_lcms._lcms_LUT_struc._fields_
    # CLUT itself is in the 'T' field
    ALLOWED_TAGS = (c_lcms.icSigAToB0Tag,
                    c_lcms.icSigAToB1Tag,
                    c_lcms.icSigAToB2Tag,
                    c_lcms.icSigBToA0Tag,
                    c_lcms.icSigBToA1Tag,
                    c_lcms.icSigBToA2Tag,
                    c_lcms.icSigGamutTag,
                    c_lcms.icSigPreview0Tag,
                    c_lcms.icSigPreview1Tag,
                    c_lcms.icSigPreview2Tag)
    SUPPORTED_TAGS = (c_lcms.icSigAToB0Tag, # LPLUT cmsReadICCLut(cmsHPROFILE hProfile, icTagSignature sig)
                      c_lcms.icSigAToB1Tag, # cmsReadICCLut
                      c_lcms.icSigAToB2Tag, # cmsReadICCLut
                      c_lcms.icSigBToA0Tag, # cmsReadICCLut
                      c_lcms.icSigBToA1Tag, # cmsReadICCLut
                      c_lcms.icSigBToA2Tag) # cmsReadICCLut

    
    def __init__(self, tagType, payload=None, profile=None):
        """
        Constructor.
        
        @param tagType: ICC tagType of tag to create.
        @tagType tagType: C{int}
        @param payload: Payload of the tag (default: None).
            This is in terms that the low level L{c_lcms} bindings can
            understand directly.
        @type payload: C{object}
        @param profile: A profile to extract the tag information from.
        @type profile: L{Profile}
        """
        Tag.__init__(self, tagType, payload, profile)
        if payload is not None:
            if isinstance(payload, LUT):
                self.payload = payload
            elif isinstance(payload, c_lcms.LPLUT):
                self.payload = LUT(lcms_lut=payload)
        elif profile is None:
            self.payload = LUT(17)
    
    
    def _content_from_profile(self, tagType, profile):
        """Extract tag information from profile."""
        Tag._content_from_profile(self, tagType, profile)
    
        my_lcms_lut = c_lcms.cmsReadICCLut(profile.lcmsProfile, tagType)
        self.payload = LUT(lcms_lut=my_lcms_lut)
        
       
    @property
    def lcms_payload(self):
        """
        Retrieve the payload for LCMS profile creation purposes.
        
        @return: Tag content as usable by lcms directly.
        @rtype: C{littlecms.LPLUT}
        """
        return self.payload.content
    
    
    @property
    def content(self):
        """
        Dereferenced pointer to lcms LUT content.
        
        @return: Tag content for ease of to access content.
        @rtype: C{littlecms.c_lcms._lcms_LUT_struc}
        """
        return self.payload.content[0]



class ColourTag(Tag):
    """
    Colour information type tags.
    """
    ALLOWED_TAGS = (c_lcms.icSigRedColorantTag,
                    c_lcms.icSigGreenColorantTag,
                    c_lcms.icSigBlueColorantTag,
                    c_lcms.icSigMediaWhitePointTag,
                    c_lcms.icSigMediaBlackPointTag)
    SUPPORTED_TAGS = (c_lcms.icSigRedColorantTag,
                      c_lcms.icSigGreenColorantTag,
                      c_lcms.icSigBlueColorantTag,
                      c_lcms.icSigMediaWhitePointTag,
                      c_lcms.icSigMediaBlackPointTag)
    _ACCESSOR_FUNCTIONS = {c_lcms.icSigRedColorantTag:     c_lcms.cmsTakeColorants,
                           c_lcms.icSigGreenColorantTag:   c_lcms.cmsTakeColorants,
                           c_lcms.icSigBlueColorantTag:    c_lcms.cmsTakeColorants,
                           c_lcms.icSigMediaWhitePointTag: c_lcms.cmsTakeMediaWhitePoint,
                           c_lcms.icSigMediaBlackPointTag: c_lcms.cmsTakeMediaBlackPoint}
                           
    
    def __init__(self, tagType, payload=None, profile=None):
        """
        Constructor.
        
        @param tagType: ICC tagType of tag to create.
        @tagType tagType: C{int}
        @param payload: Payload of the tag (default: None).
            This is in terms that the low level L{c_lcms} bindings can
            understand directly.
        @type payload: C{object}
        @param profile: A profile to extract the tag information from.
        @type profile: L{Profile}
        """
        Tag.__init__(self, tagType, payload, profile)
        if self.payload is None:
            self.payload = CieColour()
        if payload is not None:
            self.payload = CieColour(payload)
        

    def _content_from_profile(self, tagType, profile):
        """Extract tag information from profile."""
        Tag._content_from_profile(self, tagType, profile)
    
        accessor_function = self._ACCESSOR_FUNCTIONS[tagType]
        
        if tagType == c_lcms.icSigMediaWhitePointTag:
            colour = c_lcms.cmsCIEXYZ(0, 0, 0)
            accessor_function(colour, profile.lcmsProfile)
            self.payload = CieColour(colour)
        elif tagType == c_lcms.icSigMediaBlackPointTag:
            colour = c_lcms.cmsCIEXYZ(0, 0, 0)
            accessor_function(colour, profile.lcmsProfile)
            self.payload = CieColour(colour)
        else:
            colours = c_lcms.cmsCIEXYZTRIPLE()
            accessor_function(colours, profile.lcmsProfile)
            if tagType == c_lcms.icSigRedColorantTag:
                self.payload = CieColour(colours.Red)
            elif tagType == c_lcms.icSigGreenColorantTag:
                self.payload = CieColour(colours.Green)
            elif tagType == c_lcms.icSigBlueColorantTag:
                self.payload = CieColour(colours.Blue)
            else:
                raise LcmsException('This should not happen!!!')
            
       
    @property
    def lcms_payload(self):
        """
        Retrieve the payload for LCMS profile creation purposes.
        """
        return ctypes.pointer(self.payload.lcms_XYZ)



class GammaTag(Tag):
    """
    Colour information type tags.
    """
    ALLOWED_TAGS = (c_lcms.icSigRedTRCTag,
                    c_lcms.icSigGreenTRCTag,
                    c_lcms.icSigBlueTRCTag,
                    c_lcms.icSigGrayTRCTag)
    SUPPORTED_TAGS = ()
    
    
    def __init__(self, tagType, payload=None, profile=None):
        """
        Constructor.
        
        @param tagType: ICC tagType of tag to create.
        @tagType tagType: C{int}
        @param payload: Payload of the tag (default: None).
            This is in terms that the low level L{c_lcms} bindings can
            understand directly.
        @type payload: C{object}
        @param profile: A profile to extract the tag information from.
        @type profile: L{Profile}
        """
        Tag.__init__(self, tagType, payload, profile)


    @property
    def lcms_payload(self):
        """
        Retrieve the payload for LCMS profile creation purposes.
        """
        return self.payload.content



# For convenience create a tag signature to class dictionary.
_TAG_TYPES = [TextTag, ColourTag, GammaTag, LutTag]
_TAG_SIGNATURES = []
_TAG_CLASSES = []
for item in _TAG_TYPES:
    _TAG_SIGNATURES.extend(item.ALLOWED_TAGS)
    _TAG_CLASSES.extend([item] * len(item.ALLOWED_TAGS))
TAG_CLASS_MAP = dict(zip(_TAG_SIGNATURES, _TAG_CLASSES))

# Currently unsupported tag types: 
#              Tag: (
#        c_lcms.icSigChromaticityTag         LPcmsCIExyYTRIPLE
#        c_lcms.icSigNamedColor2Tag          LPcmsNAMEDCOLORLIST
#        c_lcms.icSigColorantTableTag        LPcmsNAMEDCOLORLIST
#        c_lcms.icSigColorantTableOutTag     LPcmsNAMEDCOLORLIST
#        c_lcms.icSigCalibrationDateTimeTag  const struct tm*
