#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=
'''
Handling ICC profile based colour transformation in an OO front end to LCMS bindings.
'''

## Created: 2008-10-13 Guy K. Kloss <G.Kloss@massey.ac.nz>
## Changed: $Id:$
##
## Copyright 2008 Guy K. Kloss
## All rights reserved.
##
## This software is distributed under the terms of the
## GNU Lesser General Public License
##
## http://www.massey.ac.nz/~gkloss/

__author__ = 'Guy K. Kloss <G.Kloss@massey.ac.nz>'
__date__ = '$Date: $'
__version__ = '$Revision: $'[11:-2]

import numpy
from operator import attrgetter
import PIL
from littlecms import c_lcms, general
from littlecms.general import LcmsException


def _getColourType(profile, depth, specialFormat=None):
    """
    Get the LCMS colour type from a profile and format specifications.
    """
    typeValue = None
    try:
        if depth == 0:
            depth = 'DBL'
        typeString = 'c_lcms.TYPE_%s_%s' % (profile.colourSpace.name, depth)
        if specialFormat:
            typeString += '_%s' % specialFormat
        typeValue = eval(typeString)
    except NameError:
        raise LcmsException('Invalid colour type encoding: "%s".' % typeString)
    return typeValue



class Transform(object):
    """
    Class to handle transformations between colour spaces involving ICC
    profiles.
    """
    _DEPTH_TYPES = {0: numpy.float64,
                    8: numpy.uint8,
                    16: numpy.uint16}

    
    def __init__(self, inputProfile, outputProfile,
                 renderingIntent=c_lcms.INTENT_PERCEPTUAL,
                 transformationFlags=0,
                 inputDepth=8, outputDepth=8,
                 specialInputFormat=None, specialOutputFormat=None):
        """
        The constructor, creating a transform object for colour 
        transformation.

        @param inputProfile: Input profile for conversion.
        @type inputProfile: L{Profile}
        @param outputProfile: Output profile for conversion.
        @type outputProfile: L{Profile}
        @param renderingIntent: C{int}, a valid rendering intent (default:
            L{INTENT_PERCEPTUAL}). Valid intents are L{INTENT_PERCEPTUAL},
            L{INTENT_RELATIVE_COLORIMETRIC}, L{INTENT_SATURATION} and
            L{INTENT_ABSOLUTE_COLORIMETRIC}.
        @type renderingIntent: Rendering intent constant.
        @param transformationFlags: LCMS transformation flags
            (default: 0). Refer to the LCMS API documentation at
            U{http://littlecms.com/LCMSAPI.TXT}.
        @type transformationFlags: C{int}
        @param inputDepth: Bit depth of colour channel (default: 8). Legal values
            are 8 and 16 bits, as well as 0 for double precision tuples.
        @type inputDepth: C{int} 
        @param outputDepth: Bit depth of colour channel (default: 8). Legal values
            are 8 and 16 bits, as well as 0 for double precision tuples.
        @type outputDepth: C{int}
        @param specialInputFormat: Additional format specifier (default: None).
            Legal values are None, 'SE', 'PLANAR' and 'REV'.
        @type specialInputFormat: C{string}
        @param specialOutputFormat: Additional format specifier (default: None).
            Legal values are None, 'SE', 'PLANAR' and 'REV'.
        @type specialOutputFormat: C{string}
        """
        self._myTransform = None
        self._inputProfile = inputProfile
        self._outputProfile = outputProfile
        self._renderingIntent = renderingIntent
        self._transformationFlags = transformationFlags
        if not inputDepth in [0, 8, 16]:
            raise LcmsException('Invalid input bit depth for transformation: %s.'
                                % inputDepth)
        if not outputDepth in [0, 8, 16]:
            raise LcmsException('Invalid output bit depth for transformation: %s.'
                                % outputDepth)
        self.inputDepth = inputDepth
        self.outputDepth = outputDepth
        if not specialInputFormat in [None, 'SE', 'PLANAR', 'REV']:
            raise LcmsException('Invalid additional format specifier for input: "%s".'
                                % specialInputFormat)
        if not specialOutputFormat in [None, 'SE', 'PLANAR', 'REV']:
            raise LcmsException('Invalid additional format specifier for output: "%s".'
                                % specialOutputFormat)
        self.specialInputFormat = specialInputFormat
        self.specialOutputFormat = specialOutputFormat

        # Detect the input/output format.
        self._inputFormat = _getColourType(self._inputProfile,
                                           self.inputDepth,
                                           self.specialInputFormat)
        self._outputFormat = _getColourType(self._outputProfile,
                                            self.outputDepth,
                                            self.specialOutputFormat)

        self._createTransform()


    def __del__(self):
        """
        Auto-deleting internal.
        """
        if self._myTransform:
            c_lcms.cmsDeleteTransform(self._myTransform)


    def _createTransform(self):
        """
        Create the internal structures for LCMS transform objects.
        """
        if self._myTransform:
            c_lcms.cmsDeleteTransform(self._myTransform)
            self._myTransform = None
        self._myTransform = c_lcms.cmsCreateTransform(self._inputProfile.lcmsProfile,
                                                      self._inputFormat,
                                                      self._outputProfile.lcmsProfile,
                                                      self._outputFormat,
                                                      self._renderingIntent,
                                                      self._transformationFlags)
        if self._myTransform == None:
            raise LcmsException('Error creating transform.')


    def _setInputProfile(self, aProfile):
        """
        Replace the input profile on a transform object.

        @param aProfile: Input profile for conversion.
        @type aProfile: L{Profile}
        """
        self._inputProfile = aProfile
        self._createTransform()
        
    inputProfile = property(attrgetter('_inputProfile'), _setInputProfile)

    
    def _setOutputProfile(self, aProfile):
        """
        Replace the output profile on a transform object.

        @param aProfile: Output profile for conversion.
        @type aProfile: L{Profile}
        """
        self._outputProfile = aProfile
        self._createTransform()
        
    outputProfile = property(attrgetter('_outputProfile'), _setOutputProfile)

    
    def _setTransformationFlags(self, theTransformationFlags):
        """
        Replace the output profile on a transform object.

        @param theTransformationFlags: Output profile for conversion.
        @type theTransformationFlags: L{Profile}
        """
        self._transformationFlags = theTransformationFlags
        self._createTransform()
        
    transformationFlags = property(attrgetter('_transformationFlags'),
                                   _setTransformationFlags)

    
    def _create_matching_buffer(self, in_buffer):
        """
        Creates and returns a buffer of the same type and size as the
        L{in_buffer}.
        
        @param in_buffer: Input buffer
        @type in_buffer: L{numpy.ndarray} or L{PIL.Image.Image}
        
        @return: New, matching buffer object.
        @rtype: L{numpy.ndarray} or L{PIL.Image.Image} (same as L{in_buffer}.
        """
        out_buffer = None
        if isinstance(in_buffer, numpy.ndarray):
            out_buffer = numpy.empty(in_buffer.shape,
                                     dtype=self._DEPTH_TYPES[self.outputDepth])
        elif isinstance(in_buffer, PIL.Image.Image):
            out_buffer = PIL.Image.new(in_buffer.mode, in_buffer.size, None)
        else:
            raise LcmsException('Invalid input buffer type: "%s".'
                                % type(in_buffer))
        return out_buffer
    
    
    def _transform_numpy_array(self, sourceBuffer, destinationBuffer):
        """
        Transform a L{numpy.ndarray} type buffer.
        """
        original_shape = sourceBuffer.shape
        new_shape = original_shape
        
        # Reshape our baby to only need one type of processing loop.
        if sourceBuffer.ndim == 1:
            new_shape = (1, sourceBuffer.shape[0] / general.CHANNELS, general.CHANNELS)
        elif sourceBuffer.ndim == 2:
            if sourceBuffer.shape[1] != general.CHANNELS:
                raise LcmsException('Unsupported number of colour channels in 2D input buffer: %d'
                                    % sourceBuffer.shape[1])
            new_shape = (1, sourceBuffer.shape[0], sourceBuffer.shape[1])
        elif sourceBuffer.ndim == 3:
            if sourceBuffer.shape[2] != general.CHANNELS:
                raise LcmsException('Unsupported number of colour channels in 3D input buffer: %d'
                                    % sourceBuffer.shape[1])
        else:
            raise LcmsException('Unsupported shape of input buffer: %s'
                                % sourceBuffer.shape)
        
        sourceBuffer.shape = new_shape
        destinationBuffer.shape = new_shape
        
        for i in xrange(new_shape[0]):
            c_lcms.cmsDoTransform(self._myTransform,
                                  sourceBuffer[i].ctypes,
                                  destinationBuffer[i].ctypes,
                                  new_shape[1])
            
        # Change back the shape to original dimensions.
        sourceBuffer.shape = original_shape
        destinationBuffer.shape = original_shape
    

    def _transform_PIL_image(self, sourceBuffer, destinationBuffer):
        """
        Transform a L{PIL.Image.Image} type buffer.
        """
        in_array = numpy.asarray(sourceBuffer)
        out_array = numpy.zeros(in_array.shape, in_array.dtype)
        
        # Transform the image array line by line.
        for i in xrange(in_array.shape[0]):
            c_lcms.cmsDoTransform(self._myTransform,
                                  in_array[i].ctypes,
                                  out_array[i].ctypes,
                                  in_array.shape[1])
        
        out_image = PIL.Image.fromarray(out_array)
        destinationBuffer.paste(out_image, None)


    def doTransform(self, sourceBuffer,
                    destinationBuffer=None,
                    numberTuples=None):
        """
        Perform the colour space transformation.
        
        So far transformations of native L{ctypes} arrays, L{numpy.ndarray} 
        arrays and L{PIL} images are supported. For some types of 
        L{numpy.ndarray} arrays (suitable 2 and 3 dimensional arrays) and for 
        L{PIL} images the array dimensions can be implied from the array's 
        shape. Otherwise the parameter C{numberTuples} must be given.

        @param sourceBuffer: Input buffer for conversion.
        @type sourceBuffer: L{numpy.ndarray} or L{PIL.Image.Image}
        @param destinationBuffer: Output buffer for conversion (default: None).
            If no L{oputputBuffer} is given, a suitable L{destinationBuffer} of 
            the same type and shape as the L{sourceBuffer} will be created and 
            returned.
        @type destinationBuffer: L{Profile}
        @param numberTuples: Number of colour tuples to convert (default: None).
            If it is not given, an attempt to imply it from the size of the
            L{sourceBuffer} will be made.
        @type numberTuples: C{int}

        @return: The converted L{destinationBuffer} object.
        @rtype: Type as passed in, L{numpy.ndarray} or L{PIL.Image.Image} of 
            same type
        """
        if destinationBuffer == None:
            destinationBuffer = self._create_matching_buffer(sourceBuffer)

        # Sanity checks: Are NumPy buffers of compatible type and shape?
        if isinstance(sourceBuffer, numpy.ndarray):
            if not issubclass(sourceBuffer.dtype.type,
                              self._DEPTH_TYPES[self.inputDepth]):
                raise LcmsException('Illegal source buffer type. Is %s, must be %s' %
                                    (sourceBuffer.dtype.type,
                                     self._DEPTH_TYPES[self.inputDepth]))
            if not issubclass(destinationBuffer.dtype.type,
                              self._DEPTH_TYPES[self.outputDepth]):
                raise LcmsException('Illegal destination buffer type. Is %s, must be %s' %
                                    (destinationBuffer.dtype.type,
                                     self._DEPTH_TYPES[self.outputDepth]))
            if not sourceBuffer.shape == destinationBuffer.shape:
                raise LcmsException('Shape mismatch between input and output buffer.')

        # Use different transformation strategies, depending on the buffer type.
        if isinstance(sourceBuffer, numpy.ndarray):
            self._transform_numpy_array(sourceBuffer, destinationBuffer)
        elif isinstance(sourceBuffer, PIL.Image.Image):
            self._transform_PIL_image(sourceBuffer, destinationBuffer)
        else:
            # Just pretend an ordinary ctypes array.
        
            if numberTuples == None:
                numberTuples = len(sourceBuffer) / general.CHANNELS
            
            c_lcms.cmsDoTransform(self._myTransform,
                                  sourceBuffer.ctypes,
                                  destinationBuffer.ctypes,
                                  numberTuples)
        
        # Return our result as well.
        return destinationBuffer
